/*********************************************************************************
 ** Universidade Nova de Lisboa - Faculdade de Ciências e Tecnologia
 **
 ** Protocolo MAC GDCF - mac-gdcf.cc
 **
 ** Autores: 	
 ** Luís Lopes  
 ** Rodolfo Oliveira
 **  
 ** 20 de Outubro de 2009
 *********************************************************************************/
#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"
#include "mobilenode.h"

// #define DEBUG 99

#include "arp.h"
#include "ll.h"
//#include "mac.h"

#include "mac-gdcf-timers.h"
#include "mac-gdcf.h"
#include "cmu-trace.h"

// Added by Sushmita to support event tracing
#include "agent.h"
#include "basetrace.h"


/* our backoff timer doesn't count down in idle times during a
 * frame-exchange sequence as the mac tx state isn't idle; genreally
 * these idle times are less than DIFS and won't contribute to
 * counting down the backoff period, but this could be a real
 * problem if the frame exchange ends up in a timeout! in that case,
 * i.e. if a timeout happens we've not been counting down for the
 * duration of the timeout, and in fact begin counting down only
 * DIFS after the timeout!! we lose the timeout interval - which
 * will is not the REAL case! also, the backoff timer could be NULL
 * and we could have a pending transmission which we could have
 * sent! one could argue this is an implementation artifact which
 * doesn't violate the spec.. and the timeout interval is expected
 * to be less than DIFS .. which means its not a lot of time we
 * lose.. anyway if everyone hears everyone the only reason a ack will
 * be delayed will be due to a collision => the medium won't really be
 * idle for a DIFS for this to really matter!!
 */

inline void
Mac_gdcf::checkBackoffTimer()
{
	if(is_idle() && mhBackoff_.paused())
		mhBackoff_.resume(phymib_.getDIFS());
	if(! is_idle() && mhBackoff_.busy() && ! mhBackoff_.paused())
		mhBackoff_.pause();
}

inline void
Mac_gdcf::transmit(Packet *p, double timeout)
{
	tx_active_ = 1;
	
	if (EOTtarget_) {
		assert (eotPacket_ == NULL);
		eotPacket_ = p->copy();
	}

	/*
	 * If I'm transmitting without doing CS, such as when
	 * sending an ACK, any incoming packet will be "missed"
	 * and hence, must be discarded.
	 */

	if(rx_state_ != GDCFMAC_IDLE) {
		assert(pktRx_);
		struct hdr_cmn *ch = HDR_CMN(pktRx_);
		ch->error() = 1;        /* force packet discard */
	}

	/*
	 * pass the packet on the "interface" which will in turn
	 * place the packet on the channel.
	 *
	 * NOTE: a handler is passed along so that the Network
	 *       Interface can distinguish between incoming and
	 *       outgoing packets.
	 */
	downtarget_->recv(p->copy(), this);	
	mhSend_.start(timeout);
	mhIF_.start(txtime(p));
}
inline void
Mac_gdcf::setRxState(gdcfMacState newState)
{
	rx_state_ = newState;
	checkBackoffTimer(); 
}

inline void
Mac_gdcf::setTxState(gdcfMacState newState)
{
	tx_state_ = newState;
	checkBackoffTimer();
}


/* ======================================================================
   TCL Hooks for the simulator
   ====================================================================== */
static class Mac_gdcfClass : public TclClass {
public:
	Mac_gdcfClass() : TclClass("Mac/gdcf") {
	}
	TclObject* create(int, const char*const*) {
	return (new Mac_gdcf());

}
} class_mac_gdcf;


/* ======================================================================
   Mac  and Phy MIB Class Functions
   ====================================================================== */

gdcfPHY_MIB::gdcfPHY_MIB(Mac_gdcf *parent)
{
	/*
	 * Bind the phy mib objects.  Note that these will be bound
	 * to Mac/802_11 variables
	 */
	parent->bind("CWMin_", &CWMin); 
	parent->bind("CWMax_", &CWMax); 
	parent->bind("SlotTime_", &SlotTime);
	parent->bind("SIFS_", &SIFSTime);
	
	parent->bind("PreambleLength_", &PreambleLength);
	parent->bind("PLCPHeaderLength_", &PLCPHeaderLength);
	parent->bind_bw("PLCPDataRate_", &PLCPDataRate);
}

gdcfMAC_MIB::gdcfMAC_MIB(Mac_gdcf *parent)
{
	/*
	 * Bind the phy mib objects.  Note that these will be bound
	 * to Mac/802_11 variables
	 */
}

	
/* ======================================================================
   Mac Class Functions
   ====================================================================== */
Mac_gdcf::Mac_gdcf() : 
	Mac(), phymib_(this), macmib_(this), mhIF_(this), 
	mhRecv_(this), mhSend_(this), 
	mhDefer_(this),mhBackoff_(this)
{
	retries = 7; // numero de retries
	cont_successful_Tx = 0;
	cw_ = phymib_.getCWMin(); 
	
	tx_state_ = rx_state_ = GDCFMAC_IDLE;
	tx_active_ = 0;
	eotPacket_ = NULL;

    pktCTRL_ = 0;	      
	head = 0;
	
	// chk if basic/data rates are set
	// otherwise use bandwidth_ as default;	
	Tcl& tcl = Tcl::instance();
	tcl.evalf("Mac/802_11 set basicRate_");
	if (strcmp(tcl.result(), "0") != 0) 
		bind_bw("basicRate_", &basicRate_);
	else
		basicRate_ = bandwidth_;

	tcl.evalf("Mac/802_11 set dataRate_");
	if (strcmp(tcl.result(), "0") != 0) 
		bind_bw("dataRate_", &dataRate_);
	else
		dataRate_ = bandwidth_;
	
	tcl.evalf("Mac/gdcf set c_");
	if (strcmp(tcl.result(), "0") != 0) 
		bind_bw("c_", &c_);
	else
		c_ = 4;
		
    EOTtarget_ = 0;
   	bss_id_ = IBSS_ID;       	
   	
   	log_stats = NULL;
}


int
Mac_gdcf::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "eot-target") == 0) {
			EOTtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if (EOTtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}
		if (strcmp(argv[1], "log-target") == 0) { 
			logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		} else if(strcmp(argv[1], "nodes") == 0) {
			return TCL_OK;
		}
		else if (strcmp(argv[1], "set_num_nodes") == 0) 
		{
			num_nodes = atoi(argv[2]);
		    return (TCL_OK);                        		
		}
	}
	else
	{	
		if (strcmp(argv[1], "create_queue_logger")==0)
				{
					log_stats = (logger*) new logger(num_nodes);
					printf("MAC %d-- new statistics logger instantiated\n",index_);
					return(TCL_OK);
				}
	}
	
	return Mac::command(argc, argv);
}





/* ======================================================================
   Debugging Routines
   ====================================================================== */
void
Mac_gdcf::trace_pkt(Packet *p) 
{
	struct hdr_cmn *ch = HDR_CMN(p);
	struct hdr_mac_gdcf* dh = HDR_MAC_GDCF(p);
	u_int16_t *t = (u_int16_t*) &dh->dh_fc;

	fprintf(stderr, "\t[ %2x %2x %2x %2x ] %x %s %d\n",
		*t, dh->dh_duration,
		 ETHER_ADDR(dh->dh_ra), ETHER_ADDR(dh->dh_ta),
		index_, packet_info.name(ch->ptype()), ch->size());
}

void
Mac_gdcf::dump(char *fname)
{
	fprintf(stderr,
		"\n%s --- (INDEX: %d, time: %2.9f)\n",
		fname, index_, Scheduler::instance().clock());
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */
inline int
Mac_gdcf::hdr_dst(char* hdr, int dst )
{
	struct hdr_mac_gdcf *dh = (struct hdr_mac_gdcf*) hdr;
	
       if (dst > -2) {
               if (bss_id() == ((int)IBSS_ID)) {
			STORE4BYTE(&dst, (dh->dh_ra));
		} else if ( addr() == bss_id_) {
			if ( (u_int32_t) dst == MAC_BROADCAST) {
				STORE4BYTE(&dst, (dh->dh_ra));
			} else {
				int dst_broadcast;
				dst_broadcast = MAC_BROADCAST;
	
				STORE4BYTE(&dst_broadcast, (dh->dh_ra));
				STORE4BYTE(&dst, (dh->dh_3a));
				dh->dh_fc.fc_to_ds      = 1;
				dh->dh_fc.fc_from_ds    = 1;		
			}

		} else {

			STORE4BYTE(&bss_id_, (dh->dh_ra));
                        STORE4BYTE(&dst, (dh->dh_3a));
		}
	}

       return (u_int32_t)ETHER_ADDR(dh->dh_ra);
}

inline int 
Mac_gdcf::hdr_src(char* hdr, int src )
{
	struct hdr_mac_gdcf *dh = (struct hdr_mac_gdcf*) hdr;
        if(src > -2)
               STORE4BYTE(&src, (dh->dh_ta));
        return ETHER_ADDR(dh->dh_ta);
}

inline int 
Mac_gdcf::hdr_type(char* hdr, u_int16_t type)
{
	struct hdr_mac_gdcf *dh = (struct hdr_mac_gdcf*) hdr;
	if(type)
		STORE2BYTE(&type,(dh->dh_body));
	return GET2BYTE(dh->dh_body);
}


/* ======================================================================
   Misc Routines
   ====================================================================== */
inline int
Mac_gdcf::is_idle()
{
	if(rx_state_ != GDCFMAC_IDLE)
		return 0;
	if(tx_state_ != GDCFMAC_IDLE)
		return 0;
	
	return 1;
}


void
Mac_gdcf::discard(Packet *p, const char* why)
{
	hdr_mac_gdcf* mh = HDR_MAC_GDCF(p);
	hdr_cmn *ch = HDR_CMN(p);

	/* if the rcvd pkt contains errors, a real MAC layer couldn't
	   necessarily read any data from it, so we just toss it now */
	if(ch->error() != 0) {
		Packet::free(p);
		return;
	}

	switch(mh->dh_fc.fc_type) {
	case MAC_Type_Control:
		switch(mh->dh_fc.fc_subtype) {
			/* fall through - if necessary */
		case MAC_Subtype_ACK:
			if((u_int32_t)ETHER_ADDR(mh->dh_ra) == (u_int32_t)index_) {
				drop(p, why);
				return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Control subtype\n");
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(mh->dh_fc.fc_subtype) {
		case MAC_Subtype_Data:
			if((u_int32_t)ETHER_ADDR(mh->dh_ra) == \
                           (u_int32_t)index_ ||
                          (u_int32_t)ETHER_ADDR(mh->dh_ta) == \
                           (u_int32_t)index_ ||
                          ((u_int32_t)ETHER_ADDR(mh->dh_ra) == MAC_BROADCAST && mh->dh_fc.fc_to_ds == 0)) {
                                drop(p,why);
                                return;
			}
			break;
		default:
			fprintf(stderr, "invalid MAC Data subtype\n");
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "invalid MAC type (%x)\n", mh->dh_fc.fc_type);
		trace_pkt(p);
		exit(1);
	}
	Packet::free(p);
}

void
Mac_gdcf::capture(Packet *p)
{
	/*
	 * Update the NAV so that this does not screw
	 * up carrier sense.
	 */	
	Packet::free(p);
}

void
Mac_gdcf::collision(Packet *p)
{
	switch(rx_state_) {
	case GDCFMAC_RECV:
		setRxState(GDCFMAC_COLL);
		/* fall through */
	case GDCFMAC_COLL:
		assert(pktRx_);
		assert(mhRecv_.busy());
		/*
		 *  Since a collision has occurred, figure out
		 *  which packet that caused the collision will
		 *  "last" the longest.  Make this packet,
		 *  pktRx_ and reset the Recv Timer if necessary.
		 */
		if(txtime(p) > mhRecv_.expire()) {
			mhRecv_.stop();
			discard(pktRx_, DROP_MAC_COLLISION);
			pktRx_ = p;
			mhRecv_.start(txtime(pktRx_));
		}
		else {
			discard(p, DROP_MAC_COLLISION);
		}
		break;
	default:
		assert(0);
	}
}

void
Mac_gdcf::tx_resume()
{
	double rTime;
	assert(mhSend_.busy() == 0);
	assert(mhDefer_.busy() == 0);

	if(pktCTRL_) {
		/*
		 *  Need to send a CTS or ACK.
		 */
		mhDefer_.start(phymib_.getSIFS());
	} else 
	if(pktTx_) {
		if( mhBackoff_.busy() == 0) {
			hdr_cmn *ch = HDR_CMN(pktTx_);
			struct hdr_mac_gdcf *mh = HDR_MAC_GDCF(pktTx_);
			if ( (u_int32_t) ETHER_ADDR(mh->dh_ra) == MAC_BROADCAST) {
                		// no caso de broadcast isto funciona como um pre-backoff, pois já existe um novo pacote 
				// para transmitir (pktTx_ !=0)
	        		mhBackoff_.start(cw_, is_idle());
	        	} 
			else { //unicast
				//no caso do unicast iste código é um backoff intermédio no inicio de uma nova retransmissão
	        		mhBackoff_.start(cw_, is_idle());
            		}
		}
	} 
	else if(callback_) {
		Handler *h = callback_;
		callback_ = 0;
		h->handle((Event*) 0);
	}
	setTxState(GDCFMAC_IDLE);
}

void
Mac_gdcf::rx_resume()
{
	assert(pktRx_ == 0);
	assert(mhRecv_.busy() == 0);
	setRxState(GDCFMAC_IDLE);
	
}


/* ======================================================================
   Timer Handler Routines
   ====================================================================== */

void // adicionado por luis
Mac_gdcf::backoffHandler()
{
	if(pktCTRL_) {
		assert(mhSend_.busy() || mhDefer_.busy());
		return;
	}
	if(check_pktTx() == 0)
		return;
}

void
Mac_gdcf::deferHandler()
{
	// checks if ACK frame should be transmitted
	if(check_pktCTRL() == 0)
		return;
	assert(mhBackoff_.busy() == 0);
	// checks if data frame should be transmitted
	if(check_pktTx() == 0)
		return;
}




void
Mac_gdcf::recvHandler()
{
	recv_timer();

}

void
Mac_gdcf::sendHandler()
{
	send_timer();
}


void
Mac_gdcf::txHandler()
{
	if (EOTtarget_) {
		assert(eotPacket_);
		EOTtarget_->recv(eotPacket_, (Handler *) 0);
		eotPacket_ = NULL;
	}
	tx_active_ = 0;
}



/* ======================================================================
   The "real" Timer Handler Routines
   ====================================================================== */
void
Mac_gdcf::send_timer()
{
	switch(tx_state_) {

	/*
	 * Sent DATA, but did not receive an ACK packet.
	 */
	case GDCFMAC_SEND:
		RetransmitDATA();
		break;
	/*
	 * Sent an ACK, and now ready to resume transmission.
	 */
	case GDCFMAC_ACK:
		assert(pktCTRL_);
		Packet::free(pktCTRL_); 
		pktCTRL_ = 0;
		break;
	case GDCFMAC_IDLE:
		break;
	default:
		assert(0);
	}
	tx_resume();
}


/* ======================================================================
   Outgoing Packet Routines
   ====================================================================== */
int
Mac_gdcf::check_pktCTRL()
{
	struct hdr_mac_gdcf *mh;
	double timeout;
	

	if(pktCTRL_ == 0)
		return -1;
	if( tx_state_ == GDCFMAC_ACK)
		return -1;

	mh = HDR_MAC_GDCF(pktCTRL_);
							  
	switch(mh->dh_fc.fc_subtype) {
		/*
		 * IEEE 802.11 specs, section 9.2.8
		 * Acknowledments are sent after an SIFS, without regard to
		 * the busy/idle state of the medium.
		 */
	case MAC_Subtype_ACK:		
		setTxState(GDCFMAC_ACK);
		timeout = txtime(phymib_.getACKlen(), basicRate_);
		break;
	default:
		fprintf(stderr, "check_pktCTRL:Invalid MAC Control subtype\n");
		exit(1);
	}
	transmit(pktCTRL_, timeout);
	return 0;
}



int
Mac_gdcf::check_pktTx()
{
	struct hdr_mac_gdcf *mh;
	double timeout;
	
	assert(mhBackoff_.busy() == 0); 
	
	if(pktTx_ == 0)
		return -1;

	mh = HDR_MAC_GDCF(pktTx_);
	
	
	switch(mh->dh_fc.fc_subtype) {
	case MAC_Subtype_Data:
		if(! is_idle()) {
			return 0;
		}
		setTxState(GDCFMAC_SEND);
		if((u_int32_t)ETHER_ADDR(mh->dh_ra) != MAC_BROADCAST)
                        timeout = txtime(pktTx_)
                                + DSSS_MaxPropagationDelay              // XXX
                               + phymib_.getSIFS()
                               + txtime(phymib_.getACKlen(), basicRate_)
                               + DSSS_MaxPropagationDelay;             // XXX
		else
			timeout = txtime(pktTx_);
		break;
	default:
		fprintf(stderr, "check_pktTx:Invalid MAC Control subtype\n");
		exit(1);
	}
	transmit(pktTx_, timeout);
	return 0;
}



void
Mac_gdcf::sendACK(int dst)
{
	Packet *p = Packet::alloc();
	hdr_cmn* ch = HDR_CMN(p);
	struct ack_frame *af = (struct ack_frame*)p->access(hdr_mac::offset_);

	assert(pktCTRL_ == 0);

	ch->uid() = 0;
	ch->ptype() = PT_MAC;
	// CHANGE WRT Mike's code
	ch->size() = phymib_.getACKlen();
	ch->iface() = -2;
	ch->error() = 0;
	
	bzero(af, MAC_HDR_LEN);

	af->af_fc.fc_protocol_version = MAC_ProtocolVersion;
 	af->af_fc.fc_type	= MAC_Type_Control;
 	af->af_fc.fc_subtype	= MAC_Subtype_ACK;
 	af->af_fc.fc_to_ds	= 0;
 	af->af_fc.fc_from_ds	= 0;
 	af->af_fc.fc_more_frag	= 0;
 	af->af_fc.fc_retry	= 0;
 	af->af_fc.fc_pwr_mgt	= 0;
 	af->af_fc.fc_more_data	= 0;
 	af->af_fc.fc_wep	= 0;
 	af->af_fc.fc_order	= 0;

	STORE4BYTE(&dst, (af->af_ra));

	/* store ack tx time */
 	ch->txtime() = txtime(ch->size(), basicRate_);
	
	/* calculate ack duration */
 	af->af_duration = 0;	
	
	pktCTRL_ = p;
}

void
Mac_gdcf::sendDATA(Packet *p)
{
	hdr_cmn* ch = HDR_CMN(p);
	struct hdr_mac_gdcf* dh = HDR_MAC_GDCF(p);
	assert(pktTx_ == 0);

	/*
	 * Update the MAC header
	 */
	ch->size() += phymib_.getHdrLen11();

	dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
	dh->dh_fc.fc_type       = MAC_Type_Data;
	dh->dh_fc.fc_subtype    = MAC_Subtype_Data;
	
    	if ( bss_id_ != (int)IBSS_ID ) {
		{
		dh->dh_fc.fc_to_ds	= 1;
		dh->dh_fc.fc_from_ds	= 0;
		}
	} else {
		dh->dh_fc.fc_to_ds	= 0;
		dh->dh_fc.fc_from_ds	= 0;
	}	

	dh->dh_fc.fc_more_frag  = 0;
	dh->dh_fc.fc_retry      = 0;
	dh->dh_fc.fc_pwr_mgt    = 0;
	dh->dh_fc.fc_more_data  = 0;
	dh->dh_fc.fc_wep        = 0;
	dh->dh_fc.fc_order      = 0;

	/* store data tx time */
	double now = Scheduler::instance().clock(); 
 	ch->txtime() = txtime(ch->size(), dataRate_);

	if((u_int32_t)ETHER_ADDR(dh->dh_ra) != MAC_BROADCAST) {
		/* store data tx time for unicast packets */
		ch->txtime() = txtime(ch->size(), dataRate_);
		
		dh->dh_duration = usec(txtime(phymib_.getACKlen(), basicRate_)
				       + phymib_.getSIFS());



	} else {
		/* store data tx time for broadcast packets (see 9.6) */
		ch->txtime() = txtime(ch->size(), basicRate_);
		
		dh->dh_duration = 0;
	}
	pktTx_ = p;
}

/* ======================================================================
   Retransmission Routines
   ====================================================================== */

void
Mac_gdcf::RetransmitDATA()
{
	struct hdr_cmn *ch;
	struct hdr_mac_gdcf *mh;
	u_int32_t *rcount, thresh;
	
	double now = Scheduler::instance().clock(); 
	assert(mhBackoff_.busy() == 0);

	assert(pktTx_);
	
	ch = HDR_CMN(pktTx_);
	mh = HDR_MAC_GDCF(pktTx_);

	/*
	 *  Broadcast packets don't get ACKed and therefore
	 *  are never retransmitted.
	 */
	if((u_int32_t)ETHER_ADDR(mh->dh_ra) == MAC_BROADCAST) {
		Packet::free(pktTx_); 
		pktTx_ = 0;

		/*
		 * Backoff at end of TX. (post_backoff em BROADCAST)
		 */
		
		mhBackoff_.start(cw_, is_idle());
		return;
	}

	macmib_.ACKFailureCount++;
	

    rcount = &slrc_;

	cont_successful_Tx = 0;
	(*rcount)++;
	if (*rcount > retries)
	{
		/* IEEE Spec section 9.2.3.5 says this should be greater than
		   or equal */
		macmib_.FailedCount++;
		/* tell the callback the send operation failed 
		   before discarding the packet */
		hdr_cmn *ch = HDR_CMN(pktTx_);
		if (ch->xmit_failure_) {
                        ch->size() -= phymib_.getHdrLen11();
			ch->xmit_reason_ = XMIT_REASON_ACK;
                        ch->xmit_failure_(pktTx_->copy(),
                                          ch->xmit_failure_data_);
                }

		discard(pktTx_, DROP_MAC_RETRY_COUNT_EXCEEDED); 
		pktTx_ = 0;
		*rcount = 0;
	}else
	{ 
		struct hdr_mac_gdcf *dh;
		dh = HDR_MAC_GDCF(pktTx_);
		dh->dh_fc.fc_retry = 1;
		inc_cw();
		mhBackoff_.start(cw_, is_idle()); // retransmissões de unicast
	}
	
}


/* ======================================================================
   Incoming Packet Routines
   ====================================================================== */
void
Mac_gdcf::send(Packet *p, Handler *h)
{
	double rTime;
	struct hdr_mac_gdcf* dh = HDR_MAC_GDCF(p);
	
	callback_ = h;
	sendDATA(p);

	if (is_idle() && mhDefer_.busy()==0 && mhBackoff_.busy() == 0) {
		mhBackoff_.start(cw_, is_idle()); 
	}
	
	
	/// logger
	/// regista o instante em que a trama é retirada pelo MAC
	/// neste instante a trama não foi ainda transmitida, é só retirada pelo MAC
	/// da fila de espera
	double now = Scheduler::instance().clock(); // tempo actual
	hdr_queue_msg* ph = hdr_queue_msg::access(p);// ph = apontador para os dados da estatística contida no pacote
	if (log_stats!=NULL)
	log_stats->list_nodes[this->index_].starting_to_send(ph->seq,now);


}

void
Mac_gdcf::recv(Packet *p, Handler *h)
{
	struct hdr_cmn *hdr = HDR_CMN(p);
	/*
	 * Sanity Check
	 */
	assert(initialized());

	/*
	 *  Handle outgoing packets.
	 */
	if(hdr->direction() == hdr_cmn::DOWN) {
                send(p, h);
                return;
        }
	/*
	 *  Handle incoming packets.
	 *
	 *  We just received the 1st bit of a packet on the network
	 *  interface.
	 *
	 */

	/*
	 *  If the interface is currently in transmit mode, then
	 *  it probably won't even see this packet.  However, the
	 *  "air" around me is BUSY so I need to let the packet
	 *  proceed.  Just set the error flag in the common header
	 *  to that the packet gets thrown away.
	 */
	if(tx_active_ && hdr->error() == 0) {
		hdr->error() = 1;
	}

	if(rx_state_ == GDCFMAC_IDLE) {
		setRxState(GDCFMAC_RECV);
		pktRx_ = p;
		/*
		 * Schedule the reception of this packet, in
		 * txtime seconds.
		 */			
		mhRecv_.start(txtime(p));
	} else {
		/*
		 *  If the power of the incoming packet is smaller than the
		 *  power of the packet currently being received by at least
                 *  the capture threshold, then we ignore the new packet.
		 */
		if(pktRx_->txinfo_.RxPr / p->txinfo_.RxPr >= p->txinfo_.CPThresh) {
			capture(p);
		} else {
			collision(p);
		}
	}
}

void
Mac_gdcf::recv_timer()
{
	u_int32_t src; 
	hdr_cmn *ch = HDR_CMN(pktRx_);
	hdr_mac_gdcf *mh = HDR_MAC_GDCF(pktRx_);
	u_int32_t dst = ETHER_ADDR(mh->dh_ra);
	u_int32_t ap_dst = ETHER_ADDR(mh->dh_3a);
	u_int8_t  type = mh->dh_fc.fc_type;
	u_int8_t  subtype = mh->dh_fc.fc_subtype;

	assert(pktRx_);
	assert(rx_state_ == GDCFMAC_RECV || rx_state_ == GDCFMAC_COLL);
	
	
	
        /*
         *  If the interface is in TRANSMIT mode when this packet
         *  "arrives", then I would never have seen it and should
         *  do a silent discard without adjusting the NAV.
         */
        if(tx_active_) {
                Packet::free(pktRx_);
                goto done;
        }

	/*
	 * Handle collisions.
	 */
	if(rx_state_ == GDCFMAC_COLL) {
		discard(pktRx_, DROP_MAC_COLLISION);		
		goto done;
	}

	/*
	 * Check to see if this packet was received with enough
	 * bit errors that the current level of FEC still could not
	 * fix all of the problems - ie; after FEC, the checksum still
	 * failed.
	 */
	if( ch->error() ) {
		Packet::free(pktRx_);
		goto done;
	}

	
	/*
	 * IEEE 802.11 specs, section 9.2.5.6
	 *	- update the NAV (Network Allocation Vector)
	 */
	if(dst != (u_int32_t)index_) {
	//	set_nav(mh->dh_duration);
	}

        /* tap out - */
        if (tap_ && type == MAC_Type_Data &&
            MAC_Subtype_Data == subtype ) 
		tap_->tap(pktRx_);
	

	/*
	 * Address Filtering
	 */
        // pacote unicast nao enviado para este nó    
	if(dst != (u_int32_t)index_ && dst != MAC_BROADCAST) { // verifica se o pacote é para este mac.
		/*
		 *  We don't want to log this event, so we just free
		 *  the packet instead of calling the drop routine.
		 */
		discard(pktRx_, "---");
		goto done;
	}
	
	
	if ( dst == MAC_BROADCAST && mh->dh_fc.fc_to_ds == 1 && mh->dh_fc.fc_from_ds == 1) {
		if (addr() != bss_id_) {
 			discard(pktRx_, "---");
 			goto done;
		}
 	}
	
	
	switch(type) {

	case MAC_Type_Control:
		switch(subtype) {

		case MAC_Subtype_ACK:
			recvACK(pktRx_);
			break;
		default:
			fprintf(stderr,"recvTimer2:Invalid MAC Control Subtype %x\n",
				subtype);
			exit(1);
		}
		break;
	case MAC_Type_Data:
		switch(subtype) {
		case MAC_Subtype_Data:
			recvDATA(pktRx_);
			break;
		default:
			fprintf(stderr, "recv_timer3:Invalid MAC Data Subtype %x\n",
				subtype);
			exit(1);
		}
		break;
	default:
		fprintf(stderr, "recv_timer4:Invalid MAC Type %x\n", subtype);
		exit(1);
	}
 done:
	pktRx_ = 0;
	rx_resume();

}



/*
 * txtime()	- pluck the precomputed tx time from the packet header
 */
double
Mac_gdcf::txtime(Packet *p)
{
	struct hdr_cmn *ch = HDR_CMN(p);
	double t = ch->txtime();
	if (t < 0.0) {
		drop(p, "XXX");
 		exit(1);
	}
	return t;
}

 
/*
 * txtime()	- calculate tx time for packet of size "psz" bytes 
 *		  at rate "drt" bps
 */
double
Mac_gdcf::txtime(double psz, double drt)
{

	double dsz = psz - phymib_.getPLCPhdrLen();
        int plcp_hdr = phymib_.getPLCPhdrLen() << 3;	
	int datalen = (int)dsz << 3;
	double t = (((double)plcp_hdr)/(double)phymib_.getPLCPDataRate())
                                       + (((double)datalen)/(double)drt);
	return(t);
}


void
Mac_gdcf::recvDATA(Packet *p)
{
	struct hdr_mac_gdcf *dh = HDR_MAC_GDCF(p);
	u_int32_t dst, src, size;
	struct hdr_cmn *ch = HDR_CMN(p);

	dst = ETHER_ADDR(dh->dh_ra);
	src = ETHER_ADDR(dh->dh_ta);
	size = ch->size();
	/*
	 * Adjust the MAC packet size - ie; strip
	 * off the mac header
	 */
	ch->size() -= phymib_.getHdrLen11();
	ch->num_forwards() += 1;

	/*
	 *  If we sent a CTS, clean up...
	 */
	if(dst != MAC_BROADCAST) {
		{
			/*
			 *  We did not send a CTS and there's no
			 *  room to buffer an ACK.
			 */
			if(pktCTRL_) {
				discard(p, DROP_MAC_BUSY);
				return;
			}
			sendACK(src);
			if(mhSend_.busy() == 0)
			{
				
				tx_resume();
				
			}
		}
	}
	
	/* ============================================================
	   Make/update an entry in our sequence number cache.
	   ============================================================ */

	/* Changed by Debojyoti Dutta. This upper loop of if{}else was 
	   suggested by Joerg Diederich <dieder@ibr.cs.tu-bs.de>. 
	   Changed on 19th Oct'2000 */


	/*
	 *  Pass the packet up to the link-layer.
	 *  XXX - we could schedule an event to account
	 *  for this processing delay.
	 */
	
	/* in BSS mode, if a station receives a packet via
	 * the AP, and higher layers are interested in looking
	 * at the src address, we might need to put it at
	 * the right place - lest the higher layers end up
	 * believing the AP address to be the src addr! a quick
	 * grep didn't turn up any higher layers interested in
	 * the src addr though!
	 * anyway, here if I'm the AP and the destination
	 * address (in dh_3a) isn't me, then we have to fwd
	 * the packet; we pick the real destination and set
	 * set it up for the LL; we save the real src into
	 * the dh_3a field for the 'interested in the info'
	 * receiver; we finally push the packet towards the
	 * LL to be added back to my queue - accomplish this
	 * by reversing the direction!*/
	
	
		
	if ((bss_id() == addr()) && ((u_int32_t)ETHER_ADDR(dh->dh_ra)!= MAC_BROADCAST) && ((u_int32_t)ETHER_ADDR(dh->dh_3a) != ((u_int32_t)addr())) && dh->dh_fc.fc_from_ds == 0) {
		struct hdr_cmn *ch = HDR_CMN(p);
	
				
		u_int32_t dst = ETHER_ADDR(dh->dh_3a);
		u_int32_t src = ETHER_ADDR(dh->dh_ta);
		/* if it is a broadcast pkt then send a copy up
		 * my stack also
		 */
		
		if (dst == MAC_BROADCAST) {
			uptarget_->recv(p->copy(), (Handler*) 0);
		}

		ch->next_hop() = dst;

		if (dst == MAC_BROADCAST) {
			STORE4BYTE(&src, (dh->dh_3a));
		} else {
			STORE4BYTE(&src, (dh->dh_4a));			
		}

		ch->addr_type() = NS_AF_ILINK;
		ch->direction() = hdr_cmn::DOWN;
		
	}
	
 	if ((bss_id() == addr()) && dh->dh_fc.fc_to_ds == 1 && dh->dh_fc.fc_from_ds == 1) {
 		u_int32_t dst = ETHER_ADDR(dh->dh_3a);
 		u_int32_t src = ETHER_ADDR(dh->dh_4a);

 		ch->next_hop() = dst;
 		STORE4BYTE(&src, (dh->dh_3a));
 		ch->addr_type() = NS_AF_ILINK;
 		ch->direction() = hdr_cmn::DOWN;
 	}

	uptarget_->recv(p, (Handler*) 0);
	
}


void
Mac_gdcf::recvACK(Packet *p)
{	
	if(tx_state_ != GDCFMAC_SEND) {
		discard(p, DROP_MAC_INVALID_STATE);
		return;
	}
	assert(pktTx_);

	mhSend_.stop();

	
	/*
	 * The successful reception of this ACK packet implies
	 * that our DATA transmission was successful.  Hence,
	 * we can reset the Short/Long Retry Count and the CW.
	 *
	 * need to check the size of the packet we sent that's being
	 * ACK'd, not the size of the ACK packet.
	 */
	slrc_ = 0;
	cont_successful_Tx++;
	if ( cont_successful_Tx >= c_ )
	{
		dec_cw(); 
		cont_successful_Tx = 0;
	}
	Packet::free(pktTx_); 
	pktTx_ = 0;
	/*
	 * Backoff before sending again.
 	 */
	assert(mhBackoff_.busy() == 0); 
	mhBackoff_.start(cw_, is_idle()); // funciona como post-backoff para unicast

	

done:
	if (addr() == bss_id_) {
		
	}

	tx_resume();

	mac_log(p);
}
