/*********************************************************************************
 ** Universidade Nova de Lisboa - Faculdade de Ciências e Tecnologia
 **
 ** Protocolo MAC GDCF - mac-gdcf.h
 **
 ** Autores: 	
 ** Luís Lopes  
 ** Rodolfo Oliveira
 **  
 ** 20 de Outubro de 2009
 *********************************************************************************/

#ifndef ns_mac_gdcf_h 
#define ns_mac_gdcf_h

// Added by Sushmita to support event tracing (singal@nunki.usc.edu)
#include "address.h"
#include "ip.h"

#include "mac-gdcf-timers.h"
#include "marshall.h"
#include <math.h>
#include <stddef.h>


// log the mac generic statistics
#include "../gen_mac_logger/queue_logger.h" //logger de entrada da queue



class EventTrace;

#define GET_ETHER_TYPE(x)		GET2BYTE((x))
#define SET_ETHER_TYPE(x,y)            {u_int16_t t = (y); STORE2BYTE(x,&t);}


enum gdcfMacState {
	GDCFMAC_IDLE	= 0x0000,
	GDCFMAC_RECV 	= 0x0010,
	GDCFMAC_SEND 	= 0x0100,
	GDCFMAC_ACK	= 0x0800,
	GDCFMAC_COLL	= 0x1000,
};




/* ======================================================================
   Frame Formats
   ====================================================================== */
#define	MAC_ProtocolVersion	0x00

#define MAC_Type_Control	0x01
#define MAC_Type_Data		0x02
#define MAC_Type_Reserved	0x03

#define MAC_Subtype_ACK		0x0D
#define MAC_Subtype_Data	0x00



struct frame_control {
	u_char		fc_subtype		: 4;
	u_char		fc_type			: 2;
	u_char		fc_protocol_version	: 2;

	u_char		fc_order		: 1;
	u_char		fc_wep			: 1;
	u_char		fc_more_data		: 1;
	u_char		fc_pwr_mgt		: 1;
	u_char		fc_retry		: 1;
	u_char		fc_more_frag		: 1;
	u_char		fc_from_ds		: 1;
	u_char		fc_to_ds		: 1;
};



struct ack_frame {
	struct frame_control	af_fc;
	u_int16_t		af_duration;
	u_char			af_ra[ETHER_ADDR_LEN];
	u_char			af_fcs[ETHER_FCS_LEN];
};


// XXX This header does not have its header access function because it shares
// the same header space with hdr_mac.
struct hdr_mac_gdcf {
	struct frame_control	dh_fc;
	u_int16_t		dh_duration;
	u_char          dh_ra[ETHER_ADDR_LEN]; // (Receiving addr) 
    u_char          dh_ta[ETHER_ADDR_LEN]; // (transmitting address)
    u_char          dh_3a[ETHER_ADDR_LEN];
	u_char			dh_4a[ETHER_ADDR_LEN];
	u_int16_t		dh_scontrol;
	u_char			dh_body[1]; // size of 1 for ANSI compatibility
};




/* ======================================================================
   Definitions
   ====================================================================== */

/* Must account for propagation delays added by the channel model when
 * calculating tx timeouts (as set in tcl/lan/ns-mac.tcl).
 */
#define DSSS_MaxPropagationDelay        0.000002        // 2us   XXXX

class gdcfPHY_MIB {
public:
	gdcfPHY_MIB(Mac_gdcf *parent);
	inline u_int32_t getCWMin() { return(CWMin); } 
        inline u_int32_t getCWMax() { return(CWMax); } 
	inline double getSlotTime() { return(SlotTime); }
	inline double getSIFS() { return(SIFSTime); }
	inline double getPIFS() { return(SIFSTime + SlotTime); }
	inline double getDIFS() { return(SIFSTime + 2 * SlotTime); }
	inline double getEIFS() {
		// see (802.11-1999, 9.2.10)
		return(SIFSTime + getDIFS()
                       + (8 *  getACKlen())/PLCPDataRate);
	}
	inline u_int32_t getPreambleLength() { return(PreambleLength); }
	inline double getPLCPDataRate() { return(PLCPDataRate); }
		
	inline u_int32_t getPLCPhdrLen() {
		u_int32_t header_bits = PreambleLength + PLCPHeaderLength;
		u_int32_t header_bytes = header_bits>> 3;
		return(header_bytes );
	}

	inline u_int32_t getHdrLen11() {
		return(getPLCPhdrLen() + offsetof(struct hdr_mac_gdcf, dh_body[0])
                       + ETHER_FCS_LEN);
	}
		
	inline u_int32_t getACKlen() {
		return(getPLCPhdrLen() + sizeof(struct ack_frame));
	}

 private:
 	u_int32_t	CWMin; 
	u_int32_t	CWMax; 
	double		SlotTime;
	double		SIFSTime;
	u_int32_t	PreambleLength;
	u_int32_t	PLCPHeaderLength;
	double		PLCPDataRate;
};


/*
 * IEEE 802.11 Spec, section 11.4.4.2
 *      - default values for the MAC Attributes
 */
#define MAC_FragmentationThreshold	2346		// bytes
#define MAC_MaxTransmitMSDULifetime	512		// time units
#define MAC_MaxReceiveLifetime		512		// time units




class gdcfMAC_MIB {
public:
     gdcfMAC_MIB(Mac_gdcf *parent);
private:	
public:
	u_int32_t	FailedCount;	
	u_int32_t	ACKFailureCount;
 public:
};


/* ======================================================================
   The actual 802.11 MAC class.
   ====================================================================== */
class Mac_gdcf : public Mac {
	friend class DeferTimer;
	friend class IFTimer;
	friend class RxTimer;
	friend class TxTimer;
	friend class BackoffTimer;
public:
	Mac_gdcf();
	void		recv(Packet *p, Handler *h);
	inline int	hdr_dst(char* hdr, int dst = -2);
	inline int	hdr_src(char* hdr, int src = -2);
	inline int	hdr_type(char* hdr, u_int16_t type = 0);
	
	inline int bss_id() { return bss_id_; }

public:
	void	deferHandler(void); // invoca check_pktCTRL() e check_pktTx() 
	void	recvHandler(void);  // invoca recv_timer
	void	sendHandler(void);  // invoca send_timer
	void	txHandler(void);    // ifTimer ... timer da queue
	void	backoffHandler(void);// invoca check_pktTx()
	
private:
    int	command(int argc, const char*const* argv);
	
	/*
	 * Called by the timers.
	 */
	void    recv_timer(void);
	void	send_timer(void);
	int		check_pktCTRL();
	int		check_pktTx();
		
	/*
	 * Packet Transmission Functions.
	 */
	void	send(Packet *p, Handler *h);
	void	sendACK(int dst);
	void	sendDATA(Packet *p);
	void	RetransmitDATA();

	/*
	 * Packet Reception Functions.
	 */
	void	recvACK(Packet *p);
	void	recvDATA(Packet *p);
	
	void	capture(Packet *p);
	void	collision(Packet *p);
	void	discard(Packet *p, const char* why);
	void	rx_resume(void);
	void	tx_resume(void);

	inline int	is_idle(void);

	/*
	 * Debugging Functions.
	 */
	void		trace_pkt(Packet *p);
	void		dump(char* fname);

	inline void mac_log(Packet *p) {
                logtarget_->recv(p, (Handler*) 0);
        }

	double txtime(Packet *p);
	double txtime(double psz, double drt);
	double txtime(int bytes) { /* clobber inherited txtime() */ abort(); return 0;}

	inline void transmit(Packet *p, double timeout);
	inline void setRxState(gdcfMacState newState);
	inline void setTxState(gdcfMacState newState);
	inline void checkBackoffTimer(void); 
	
	inline void inc_cw() { 
		cw_ = (cw_ << 1) + 1;
		if(cw_ > phymib_.getCWMax())
			cw_ = phymib_.getCWMax();
	}
	inline void dec_cw() { 
		cw_ = cw_/2;
		if(cw_ < phymib_.getCWMin())
			cw_ = phymib_.getCWMin(); 
	}
	inline double sec(double t) { return(t *= 1.0e-6); }
	inline u_int16_t usec(double t) {
		u_int16_t us = (u_int16_t)floor((t *= 1e6) + 0.5);
		return us;
	}

public:
	    gdcfPHY_MIB         phymib_;
        gdcfMAC_MIB         macmib_;

       /* the macaddr of my AP in BSS mode; for IBSS mode
        * this is set to a reserved value IBSS_ID - the
        * MAC_BROADCAST reserved value can be used for this
        * purpose
        */
       int     bss_id_;
       enum    {IBSS_ID=MAC_BROADCAST};
       enum    {
    	   	PASSIVE = 0,
       		ACTIVE = 1
		};

private:
	double		basicRate_;
 	double		dataRate_;
 	double		c_;
	int head;
	/*
	 * Mac Timers
	 */
	gdcfIFTimer		mhIF_;		// interface timer
	gdcfRxTimer		mhRecv_;	// incoming packets
	gdcfTxTimer		mhSend_;	// outgoing packets

	gdcfDeferTimer		mhDefer_;	// defer timer
	gdcfBackoffTimer	mhBackoff_; 
	
/* ============================================================
	   Internal MAC State
 ============================================================ */
	gdcfMacState	rx_state_;	// incoming state (MAC_RECV or MAC_IDLE)
	gdcfMacState	tx_state_;	// outgoint state
	int			tx_active_;	// transmitter is ACTIVE

	/* ============================================================
		   Internal MAC State
	 ============================================================ */	
	Packet      *eotPacket_;    // copy for eot callback
	Packet		*pktCTRL_;	// outgoing non-RTS packet
	
	u_int32_t	cw_;		// Contention Window 
	
	int			min_frame_len_;

	NsObject*	logtarget_;
	NsObject*	EOTtarget_;     // given a copy of packet at TX end
	
	u_int32_t	slrc_;
	u_int32_t retries;
	int cont_successful_Tx;
	
	
	logger *log_stats;
	int    num_nodes; //nós da simulação necessários para saber a dimensão da tabela do logger
};

#endif /* __mac_80211_h__ */
